#  Copyright 2022 Exactpro (Exactpro Systems Limited)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations
from abc import ABC, abstractmethod

from th2_data_services.interfaces.command import ICommand
from th2_data_services.interfaces.source_api import ISourceAPI

#LOG import logging

#LOG logger = logging.getLogger(__name__)


class IDataSource(ABC):
    @abstractmethod
    def command(self, cmd: ICommand):
        pass

    @property
    @abstractmethod
    def source_api(self) -> ISourceAPI:
        pass
